package log

import (
	"testing"

	"github.com/golang/protobuf/ptypes/timestamp"
	"github.com/stretchr/testify/require"
	pb "gitlab.com/gitlab-org/gitaly-proto/go"
	"gitlab.com/gitlab-org/gitaly/internal/git/catfile"
)

func TestParseRawCommit(t *testing.T) {
	info := &catfile.ObjectInfo{
		Oid:  "a984dfa4dee018c6d5f5f57ffec0d0e22763df16",
		Type: "commit",
	}

	// Valid-but-interesting commits should be test at the FindCommit level.
	// Invalid objects (that Git would complain about during fsck) can be
	// tested here.
	//
	// Once a repository contains a pathological object it can be hard to get
	// rid of it. Because of this I think it's nicer to ignore such objects
	// than to throw hard errors.
	testCases := []struct {
		desc string
		in   []byte
		out  *pb.GitCommit
	}{
		{
			desc: "empty commit object",
			in:   []byte{},
			out:  &pb.GitCommit{Id: info.Oid},
		},
		{
			desc: "no email",
			in:   []byte("author Jane Doe"),
			out: &pb.GitCommit{
				Id:     info.Oid,
				Author: &pb.CommitAuthor{Name: []byte("Jane Doe")},
			},
		},
		{
			desc: "unmatched <",
			in:   []byte("author Jane Doe <janedoe@example.com"),
			out: &pb.GitCommit{
				Id:     info.Oid,
				Author: &pb.CommitAuthor{Name: []byte("Jane Doe")},
			},
		},
		{
			desc: "unmatched >",
			in:   []byte("author Jane Doe janedoe@example.com>"),
			out: &pb.GitCommit{
				Id:     info.Oid,
				Author: &pb.CommitAuthor{Name: []byte("Jane Doe janedoe@example.com>")},
			},
		},
		{
			desc: "missing date",
			in:   []byte("author Jane Doe <janedoe@example.com> "),
			out: &pb.GitCommit{
				Id:     info.Oid,
				Author: &pb.CommitAuthor{Name: []byte("Jane Doe"), Email: []byte("janedoe@example.com")},
			},
		},
		{
			desc: "date too high",
			in:   []byte("author Jane Doe <janedoe@example.com> 9007199254740993 +0200"),
			out: &pb.GitCommit{
				Id: info.Oid,
				Author: &pb.CommitAuthor{
					Name:  []byte("Jane Doe"),
					Email: []byte("janedoe@example.com"),
					Date:  &timestamp.Timestamp{Seconds: 9223371974719179007},
				},
			},
		},
		{
			desc: "date negative",
			in:   []byte("author Jane Doe <janedoe@example.com> -1 +0200"),
			out: &pb.GitCommit{
				Id: info.Oid,
				Author: &pb.CommitAuthor{
					Name:  []byte("Jane Doe"),
					Email: []byte("janedoe@example.com"),
					Date:  &timestamp.Timestamp{Seconds: 9223371974719179007},
				},
			},
		},
	}

	for _, tc := range testCases {
		t.Run(tc.desc, func(t *testing.T) {
			info.Size = int64(len(tc.in))
			out, err := parseRawCommit(tc.in, info)
			require.NoError(t, err, "parse error")
			require.Equal(t, *tc.out, *out)
		})
	}
}
